#!/usr/bin/env python
#
# the elliptic curve factoring method using homogeneous coordinates
#

import os
import sys
import random
import math


# - - - - - - - class definitions  - - - - - - - - - -
class curve:
	#  - - - - - - - - - - constructor
	def __init__(self,a=0,b=0,p=0):
		self.a=a;
		self.b=b;
		self.p=p;
		self.n=0;

	#  - - - - - - - - - - valid curve?
	def valid_parameters(self):
		b=self.b
		a=self.a
		p=self.p
		if (4*a*a*a+27*b*b)%p==0:
			return False;
		return True;

	#  - - - - - - - - - - compute |E| by evaluating (x^3+ax+b)^((p-1)2)
	def compute_n(self):
		self.n=0;
		x=0;
		while x<=self.p-1:
			z=(x*x*x+self.a*x+self.b)%self.p;
			if z==0:
				self.n=self.n+1
#				print "1 point for x=",x
			else:
				plus_minus_one=pow(z,(p-1)/2,p)
				if plus_minus_one==1:
					self.n=self.n+2
#					print "2 points for x=",x
			x=x+1

		# special point O
#		print "special point"
		self.n=self.n+1
		return self.n

class point:

	# - - - - - - - - - - constructor
	def __init__(self,x=0,y=0,z=0):
		self.x=x;
		self.y=y;
		self.z=z;
		if z==0:
			self.infinity=True;
		else:
			self.infinity=False;

	# - - - - - - - - - - check whether y^2*z = x^3 + a*x*z^2+b*z^3

	def is_on_curve(self,E):

		if (self.infinity==True):
			return True;

		x=self.x;
		y=self.y;
		z=self.z

		y_side= (y*y*z) % E.p;
		x_side= (x*x*x+E.a*x*z*z+b*z*z*z) % E.p;

		if y_side == x_side:
			return True;
		else:
			return False;
			
# - - - - - - - functions  - - - - - - - - - - - - - -

# - - - - - - - - - - generating small primes, sieve of Erathostenes
# returns a list of primes <= bound

def erathostenes(bound):
	
	a=[];
	for i in range(0,bound+1):
		a.append(0);
	
	a[1]=1
	i=1

	while i*i<=bound:
		while a[i]==1:
			i=i+1
		j=2*i
		while j<=bound:
			a[j]=1
			j=j+i
		i=i+1

	primes=[]
	i=1
	while i<=bound:
		if a[i]==0:
			primes.append(i)
		i=i+1

	return primes
	

# - - - - - - - - - -
#
# elliptic curve addition, the affine case, coordinates (x,y,z)  
# with z==1
#
# returns P3=P1+P2 or a divisor of n, if division does not work
# - - - - - - - - - -
def add_affine(P1,P2,E):

	P3=point()
# special P1=O or P2=O
	if P1.infinity==True:
		return P2, 0;
	if P2.infinity==True:
		return P1, 0;
# special P1=-P2
	if P1.x == P2.x and P1.y == -P2.y % E.p:
		P3.infinity=True;
		return P3, 0

# P1 == P2 -> m=derivation at (x,y)
	if P1.x == P2.x:
		x=P1.x
		(y_inv,inverted)=invert(2*P1.y,E.p)
		if inverted:
			m=((3*x*x+E.a)*y_inv) % E.p
		else:
			return point(0,0,1), y_inv;
	else:
# P1 != P2 -> m=dy/dx
		(x_inv,inverted)=invert(P2.x-P1.x,E.p)
		if inverted:
			m=((P2.y-P1.y)*x_inv) % E.p
		else:
			return point(0,0,1), x_inv;

	P3.x=(m*m-P1.x-P2.x) % E.p
	P3.y=(m*(P1.x-P3.x)-P1.y) % E.p
	P3.z=1;
	P3.infinity=False;

	return P3, 0;


# X3 = v * (Z2 * (Z1 * u^2 - 2 * X1 * v^2) - v^3) mod p
# 
# Y3 = Z2 * (3 * X1 * u * v^2 - Y1 * v^3 - Z1 * u^3) + u * v^3 mod p
# 
# Z3 = v^3 * Z1 * Z2 mod p
# 
# where u = Y2 * Z1 - Y1 * Z2 mod p and v = X2 * Z1 - X1 * Z2 mod p.
# 
# When the points P1 and P2 are equal, then (X1/Z1, Y1/Z1) is equal to
# (X2/Z2, Y2/Z2), which is true if and only if u and v are both equal
# to zero.
# 
# The product P3=(X3,Y3,Z3) = P1 * P1 is given by
# 
# X3 = 2 * Y1 * Z1 * (w^2 - 8 * X1 * Y1^2 * Z1) mod p
# 
# Y3 = 4 * Y1^2 * Z1 * (3 * w * X1 - 2 * Y1^2 * Z1) - w^3 mod p
# 
# Z3 = 8 * (Y1 * Z1)^3 mod p
# 
# where w = 3 * X1^2 + a * Z1^2 mod p.  In the above equations, a, u,
# v, w, X1, X2, X3, Y1, Y2, Y3, Z1, Z2, and Z3 are integers in the set
# Fp.
# 

#
# P3=P1+P2 with projective coordinates (x,y,z) method taken
# from https://tools.ietf.org/rfc/rfc6090.txt
#

def add(P1,P2,E):

# special P1=O or P2=O
	if P1.infinity==True:
		return P2;
	if P2.infinity==True:
		return P1;

# auxiliary variables to speed up the computation

	P1xP2z=P1.x*P2.z;
	P2xP1z=P2.x*P1.z;
	P2xP1z_P1xP2z=(P2xP1z-P1xP2z) % E.p
	P2yP1z=P2.y*P1.z;
	P2zP1y=P2.z*P1.y;

	xz_equal=(P2xP1z_P1xP2z == 0)
        yz_neg_equal=(((P2yP1z+P2zP1y) % E.p) == 0)
	if xz_equal and yz_neg_equal:
		P3=point(0,1,0); # infinity
#		print "# ret P3 = ",
#		output_point(P3);
		return P3;

	P3=point() # storing (x,y,z)-coordinates of result

	u=(P2yP1z-P2zP1y) % E.p
	v=P2xP1z_P1xP2z

	if u==0 and v==0: # ........................P1 = P2
		w=(3*P1.x*P1.x+E.a*P1.z*P1.z) % E.p;
		P1yP1z=(P1.y*P1.z) % E.p
		P1yP1yP1z=(P1.y*P1yP1z) % E.p
		ww=(w*w) % E.p
		P3.x=(2*P1yP1z*(ww-8*P1.x*P1yP1yP1z)) % E.p
		P3.y=(4*P1yP1yP1z*(3*w*P1.x-2*P1yP1yP1z)-ww*w) % E.p
		P3.z=(8*P1yP1z*P1yP1z*P1yP1z) % E.p;
	else: 	# ..................................P1 != P2
		uu=(u*u) % E.p
		uuu=(uu*u) % E.p
		vv=(v*v) % E.p
		vvv=(vv*v) % E.p
		P3.x=(v*(P2.z*(P1.z*uu-2*P1.x*vv)-vvv)) % E.p;
		P3.y=(P2.z*(3*P1.x*u*vv-P1.y*vvv-P1.z*uuu) + u*vvv) % E.p;
		P3.z=(vvv*P1.z*P2.z) % E.p;

	

	P3.infinity=False;


	return P3;


# gcd() by Euclid's algorithm

def gcd(a,b):
	r=1;
	while r>0:
		q=a/b;
		r=a%b;
		a=b;
		b=r
	return a

# extended Euclid's algorithm
# returns d=gcd(a,b) and (x,y) with d=a*x+b*y
def egcd(a,b):
	r=1;
	x=1;y=0;x1=0;y1=1;
	while r>0:
		q=a/b;
		r=a%b;
		a=b;
		b=r
		x2=x
		x=x1
		x1=x2-q*x1
		y2=y
		y=y1
		y1=y2-q*y1
	return (a,x,y) # a is now d, the gcd of the input

# 
# x^(-1) mod p
# return a pair (d,boolean)
# if the boolean=true,  d is x^(-1)
# if the boolean=false, d is a factor of p
def invert(x,p):
	(a,u,v)=egcd(x,p);

	if a!=1 :
		return (a,False);
	if u<0:
		u=u+p;

	return (u,True);

# print (x,y,z)
def output_point(P):
	if P.infinity==False:
		print "(",P.x,P.y,P.z,")"
	else:
		print "O"

# simple primality test via 2^((p-1)/2) and 3^((p-1)/2)

def quick_prime_check(p):
	plus_minus_one=pow(2,(p-1)/2,p)
	if (plus_minus_one==1 or plus_minus_one==p-1):
		plus_minus_one=pow(3,(p-1)/2,p);
		if (plus_minus_one==1 or plus_minus_one==p-1):
			return True;
	return False;

# elliptic curve mult()
def mult(a,G,E):

	sum=point(0,1,0) # infinity
	D=G
	while a>0:
		a0=a % 2
		if a0==1:
			sum=add(sum,D,E)
		D=add(D,D,E)
		a=(a-a0)/2

	return sum

# - - - - - - - main  - - - - - - - - - - - - - -

# the elliptic curve factoring method
# call  ./ecm-h-ueb.py n

# defaults for 5,10,15,... decimal digits
# according to
# http://www.mersennewiki.org/
#
B1_defaults=[100, 150,300,2200,10000,25000,100000,300000,
             11000000,43000000,110000000,260000000,850000000,
	     29000000000,7600000000,25000000000];
N_CURVES_defaults=[10,20,30,86,214,430,910,2351,4482,7557,17884,
                   42057,69471,102212,188056,265557];

if len(sys.argv) <= 1:
	print "usage:",sys.argv[0]," <n>"
	exit(1);

print_threshold=100
stage1_bound=10

n=int(sys.argv[1]);
if quick_prime_check(n):
	print "is a prime n=",n;
	exit(1);


n_decimal_digits=int(math.log(n)/math.log(10.0))
if n_decimal_digits>80:
	n_decimal_digits=80;

print "searching parameters for #digits=",n_decimal_digits

while (n_decimal_digits%5!=0):
	n_decimal_digits=n_decimal_digits+1

print "    using parameters for #digits=",n_decimal_digits

B1=B1_defaults[n_decimal_digits/5-1];
n_curves=N_CURVES_defaults[n_decimal_digits/5-1];

B2=B1*10;

# generate list of small primes up to B2 (which is >B1)
primes=erathostenes(B2)

print "parameters: B1 =",B1
print "parameters: B2 =",B2
print "#curves        =",n_curves

B=1;

pti=0;
pti_max=len(primes)-1;

print "#primes up to B2: ",pti_max+1;

pp=primes[0];

# phase 1: multiply all prime powers up to B1

compute_powers=True;

while pp<=B1:
	pp_power=pp;

	j=1; # which power?
	while compute_powers: 
		pp_power_pp=pp_power*pp;
		if pp_power_pp<B1:
			pp_power=pp_power_pp;
			j=j+1;
		else:
			if j==1:
				compute_powers=False;
			break;
	B=B*pp_power;
	pti=pti+1;
	pp=primes[pti]

curves=1;

# initialize random number generator
random.seed();

while curves<=n_curves:
	x=random.randint(0,n)
	y=random.randint(0,n)
	a=random.randint(0,n)

	# compute point with affine coordinates and set z=1
	b=(y*y-x*x*x-a*x)%n;
	E=curve(a,b,n)

	print "curve:",E.a,E.b;
	if not E.valid_parameters():
		print "bad curve"
		curves=curves+1;
		continue;
	
	# terminology of 2006 CHES paper of
	# Implementing the Elliptic Curve Method of Factoring 
	# in Reconfigurable Hardware (Kwon, Baier, Kohlbrenner...)

	
	# P0 has projective coordinates
	P0=point(x,y,1)

	Q0=mult(B,P0,E)

	# print P0, Q0 once
	if curves==1:
		print "P0"
		output_point(P0);
		print "Q0"
		output_point(Q0);

	d=gcd(Q0.z,n)

	if d==n:
		print "too bad, gcd(z,n)=n"
		curves=curves+1;
		continue;

	if d!=1:
		print "stage 1 successful (Q0.z)"
		print "factor found:",d,"curve",curves
		break;

	# phase 2
	
	d=1;
		
	pp=primes[pti]
		
	while pti<pti_max: # as long as there are primes in array
	
		Q1=mult(pp,Q0,E) # try factor pp

		d=(d*Q1.z) % n;
		pti=pti+1
		pp=primes[pti]

	dd=gcd(d,n);

	if dd!=1:
		print "stage 2 successful (Q1.z multiplied)"
		print "factor found:",dd,"curve",curves
		break;
	else:
		print "sorry, ECM failed"

	curves=curves+1;

exit(0);
